package ava.parser

import ava.parser.TokenDelimiter.isWhiteSpace
import scala.annotation.tailrec
import scala.collection.mutable.ListBuffer
import scala.collection.mutable.Stack

class Tokenizer(private val reader: CharacterReader):
  import Tokenizer.*

  private val buffer: ListBuffer[Char]  = ListBuffer.empty
  private val states: Stack[State]      = Stack.empty
  private val errors: ListBuffer[Error] = ListBuffer.empty
  private var whiteSpaceOnly: Boolean   = true

  private def dumpStack(): List[State] = states.toList

  /** Consume the next available token.
    *
    * @return
    *   The next available token, or an error if resolving a token fails.
    */
  def next(): Either[Error, Token] =
    resetState()
    nextInternal(State.Initial)

  /** Close this Tokenizer instance and free all resources, including the
    * [[CharacterReader]].
    */
  def close(): Unit = reader.close()

  private def resetState(): Unit =
    buffer.clear()
    states.clear()
    errors.clear()

  @tailrec
  private def nextInternal(state: State): Either[Error, Token] =
    state match
      case State.Initial =>
        reader.consume() match
          case None => Right(Token.Eof)
          case Some(ch) =>
            ch match
              case '\n' =>
                // Every newline resets our white space check for comments.
                whiteSpaceOnly = true
                nextInternal(state)
              case _ if isWhiteSpace(ch) => nextInternal(state)
              case TokenDelimiter.OpenParen =>
                whiteSpaceOnly = false
                Right(Token.OpenParen)
              case TokenDelimiter.CloseParen =>
                whiteSpaceOnly = false
                Right(Token.CloseParen)
              case TokenDelimiter.Comma =>
                whiteSpaceOnly = false
                Right(Token.Comma)
              case TokenDelimiter.Colon =>
                whiteSpaceOnly = false
                Right(Token.Colon)
              case TokenDelimiter.Dot =>
                whiteSpaceOnly = false
                Right(Token.Dot)
              case TokenDelimiter.Tuple =>
                whiteSpaceOnly = false
                Right(Token.Tuple)
              case TokenDelimiter.BackSlash =>
                whiteSpaceOnly = false
                Left(Error.BackSlashNotAllowed(reader.currentSourcePosition()))
              case TokenDelimiter.DoubleQuote =>
                whiteSpaceOnly = false
                val st = State.InDoubleQuote(reader.currentSourcePosition())
                states.push(st)
                nextInternal(st)
              case TokenDelimiter.SingleQuote =>
                whiteSpaceOnly = false
                val st = State.InSingleQuote(reader.currentSourcePosition())
                states.push(st)
                nextInternal(st)
              case '-' =>
                if whiteSpaceOnly then
                  whiteSpaceOnly = false
                  nextInternal(
                    State.PotentialComment(reader.currentSourcePosition())
                  )
                else
                  whiteSpaceOnly = false
                  buffer.addOne(ch)
                  val st = State.InGeneric(reader.currentSourcePosition())
                  states.push(st)
                  nextInternal(st)
              case _ =>
                whiteSpaceOnly = false
                buffer.addOne(ch)
                val st = State.InGeneric(reader.currentSourcePosition())
                states.push(st)
                nextInternal(st)
      case State.PotentialComment(startPos) =>
        reader.consume() match
          case None =>
            Right(Token.Generic("-"))
          case Some(ch) =>
            ch match
              case '-' =>
                // Confirmed that this is a comment. Switch to the comment state
                // so that we can read the remainder of the line.
                states.push(state)
                nextInternal(State.InComment(reader.currentSourcePosition()))
              case '\n' =>
                // This token is only a `-` character and has been delimited by
                // a newline. Reset our white space check and return the token.
                whiteSpaceOnly = true
                Right(Token.Generic("-"))
              case ' ' | '\t' =>
                // This token is only a `-` character and has been delimited by
                // whitespace on either side. Return the token.
                whiteSpaceOnly = false
                Right(Token.Generic("-"))
              case _ =>
                // This is some generic token that starts with `-`. Switch
                // states now that we're sure about what we're reading.
                whiteSpaceOnly = false
                buffer.addOne('-')
                buffer.addOne(ch)
                val st = State.InGeneric(startPos)
                states.push(st)
                nextInternal(st)
      case State.InComment(startPos) =>
        reader.consume() match
          case None =>
            // Reaching EOF during a comment is perfectly fine.
            Right(Token.Comment(buffer.mkString.trim()))
          case Some(ch) =>
            ch match
              case '\n' =>
                // Newlines terminate a comment.
                whiteSpaceOnly = true
                Right(Token.Comment(buffer.mkString.trim()))
              case _ =>
                // Any non-newline character is considered part of a comment.
                buffer.addOne(ch)
                nextInternal(state)
      case State.InDoubleQuote(startPos) =>
        reader.consume() match
          case None => Left(Error.PrematureEof(dumpStack()))
          case Some(ch) =>
            ch match
              case '\n' =>
                // Literal Newlines are not allowed within string literals.
                Left(
                  Error.UnexpectedNewLine(reader.currentSourcePosition(), state)
                )
              case '\\' =>
                // Character escapes are supported within string literals.
                val st = State.InCharEscape(reader.currentSourcePosition())
                states.push(st)
                nextInternal(st)
              case '"' =>
                // This string literal is now closed. If anything failed inside
                // this literal, suppress those errors and return them as part
                // of the token so that parsing may continue.
                Right(
                  Token.StringLiteral(
                    value = buffer.mkString,
                    errors = errors.toList
                  )
                )
              case _ =>
                // Continue accumulating characters.
                buffer.addOne(ch)
                nextInternal(state)
      case State.InSingleQuote(startPos) =>
        reader.consume() match
          case None => Left(Error.PrematureEof(dumpStack()))
          case Some(ch) =>
            ch match
              case '\n' =>
                // Literal Newlines are not allowed within character literals.
                Left(
                  Error.UnexpectedNewLine(reader.currentSourcePosition(), state)
                )
              case '\\' =>
                // Character escapes are supported within character literals.
                val st = State.InCharEscape(reader.currentSourcePosition())
                states.push(st)
                nextInternal(st)
              case '\'' =>
                // This character literal is now closed.
                createCharacterLiteral(startPos)
              case _ =>
                // Continue accumulating characters.
                buffer.addOne(ch)
                nextInternal(state)
      case State.InCharEscape(startPos) =>
        reader.consume() match
          case None =>
            Left(Error.PrematureEof(dumpStack()))
          case Some(ch) =>
            resolveCharEscape(ch) match
              case Left(error) =>
                // Character escapes always live within some other state.
                // Capture the error, but don't just stop consuming -- pop the
                // state and return to wherever (string or character literal)
                // this error originated.
                errors.addOne(error)
                val _ = states.pop()
                nextInternal(states.top)
              case Right(actual) =>
                // Add the resolved character to the buffer and return to the
                // parent state.
                val _ = buffer.addOne(actual)
                val _ = states.pop()
                nextInternal(states.top)
      case State.InGeneric(startPos) =>
        // PEEK here! This allows us to react to the character in case it's
        // something like a delimter, allowing it to be naturally consumed
        // later.
        reader.peek() match
          case None =>
            // EOF is permitted for tokens - it forcibly terminates them.
            Right(Token.Generic(buffer.mkString))
          case Some(ch) =>
            if ch == '\n' then
              // New lines close the token AND reset our white space state.
              val _ = reader.consume()
              whiteSpaceOnly = true
              Right(Token.Generic(buffer.mkString))
            else if TokenDelimiter.isDelimiter(ch) then
              // Any delimiter forcibly terminates a token.
              // Do not consume the character so it gets picked up later.
              Right(Token.Generic(buffer.mkString))
            else
              // Non-delimiter characters are added to the token.
              val _ = reader.consume()
              buffer.addOne(ch)
              nextInternal(state)

  private def resolveCharEscape(ch: Char): Either[Error, Char] =
    CharEscape.resolve(ch) match
      case None =>
        Left(Error.InvalidCharEscape(reader.currentSourcePosition(), ch))
      case Some(escape) =>
        Right(escape.output)

  private def createCharacterLiteral(pos: SourcePosition)
    : Either[Error, Token] =
    val dump = buffer.mkString
    if dump.length() > 1 then
      Left(
        Error.MultipleCharactersInLiteral(
          sourcePosition = pos,
          candidate = dump,
          errors = errors.toList
        )
      )
    else if dump.length() < 1 then
      Left(
        Error.NoCharactersInLiteral(
          sourcePosition = pos,
          errors = errors.toList
        )
      )
    else
      val out = dump.charAt(0)
      Right(Token.CharacterLiteral(value = out, errors = errors.toList))

object Tokenizer:

  sealed trait State

  object State:

    /** The initial state. This is always the state of the [[Tokenizer]] when it
      * starts, as well as between each token.
      */
    case object Initial extends State

    case class PotentialComment(start: SourcePosition) extends State
    case class InComment(start: SourcePosition)        extends State
    case class InDoubleQuote(start: SourcePosition)    extends State
    case class InSingleQuote(start: SourcePosition)    extends State
    case class InCharEscape(start: SourcePosition)     extends State
    case class InGeneric(start: SourcePosition)        extends State

    given CanEqual[State, State] = CanEqual.derived

  end State

  sealed trait Error

  object Error:

    sealed trait Positional extends Error:
      def sourcePosition: SourcePosition

    case object NotImplemented extends Error

    /** This error occurs when the end of file is reached if the tokenizer is
      * still expecting more characters.
      *
      * @param stack
      *   The stack of states at time of error. Helps communicate what
      *   additional characters were expected.
      */
    case class PrematureEof(stack: List[State]) extends Error

    case class UnexpectedNewLine(
      sourcePosition: SourcePosition,
      currentState: State
    ) extends Positional

    /** This error occurs when the source contains a character escape indicated
      * by a backslash character within some string/character literal, but the
      * escape type is not recognized.
      *
      * @param sourcePosition
      *   The [[SourcePosition]] where the error triggered.
      * @param candidate
      *   The escape name that was attempted.
      */
    case class InvalidCharEscape(
      sourcePosition: SourcePosition,
      candidate: Char
    ) extends Positional

    /** This error occurs when a character literal is expected, but more than
      * one character is present in the accumulated buffer.
      *
      * @param sourcePosition
      *   The [[SourcePosition]] where the error triggered.
      * @param candidate
      *   The candidate string.
      */
    case class MultipleCharactersInLiteral(
      sourcePosition: SourcePosition,
      candidate: String,
      errors: List[Error]
    ) extends Error

    case class NoCharactersInLiteral(
      sourcePosition: SourcePosition,
      errors: List[Error]
    ) extends Error

    case class BackSlashNotAllowed(
      sourcePosition: SourcePosition
    ) extends Error

  end Error

end Tokenizer
